#!/bin/sh
#
# Automate the building of the snapshot
# to do it manually each time.
#
#  This script expects to be executed in the top directory of the 
#  pilot-db project.  It builds three zip files:
#
#
#  snapshot-src.zip     - a copy of the current source code
#  snapshot-prc.zip     - a copy of the prc files and minimal documents
#  snapshot-plugins.zip - all of the plugins
#
#  Usage:   ./build-snapshot <language> <configure flags>
#
#  e.g:   ./build-snapshot en
#
#
#



# Get the language

if [ "x$1" = "x" ]; then
	lang="en"
else
	lang="$1"
fi


#  get rid of the first argument... now $@ refers to what is unprocessed
shift 1

die() {
	echo $1
	exit
}


#
# Make a release directory structure based on the name
#
prc_dir="snapshot-${lang}-prc"
plugin_dir="snapshot-${lang}-plugins"

prc_zip="../${prc_dir}.zip"
plugin_zip="../${plugin_dir}.zip"
src_zip="snapshot-src.zip"


#
# Make sure we aren't going to overwrite anything
#
[ ! -e $prc_dir ] || die "The directory ${prc_dir} exists."
[ ! -e $prc_dir ] || die "The directory ${plugin_dir} exists."
[ ! -e ${prc_zip} ] || die "The prc file ${prc_zip} exists."
[ ! -e ${plugin_zip} ] || die "The plugin file ${plugin_zip} exists."
[ ! -e "../${src_zip}" ] || die "The source file  ../${src_zip} exists."


mkdir $prc_dir
mkdir $plugin_dir


#
#  define some potential versions of the db application
#
light="--enable-light"
mid="--enable-mid"
standard="--enable-standard"


#
#  Here is a list of the builds that go into the snapshot-prc
#
builds="light mid standard"

#
#  Files and directories to be copied into the snapshot-prc
#
files="NEWS ChangeLog README*"




# Do some initial cleanups.
if [ -f Makefile ]; then
	make distclean 
	[ $? == 0 ] || die "couldn't clean the distribution"
fi


# some temp files
hfile=${prc_dir}/__HEAD__
bfile=${prc_dir}/__BIX__


# make all the builds defined above
for build in ${builds}; do

	if [ -f Makefile ]; then
		make clean
		[ $? == 0 ] || die "couldn't clean the build"
	fi
	echo "++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++"
	echo "+ Making ${build}"
	echo "+ Options: ${!build} $@"
	echo "++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++"

	# Pass along any extra arguments sent to this script...
	./configure ${!build} $@
	make 
	[ $? == 0 ] || die " **  Build Failed!  (db-${build}) ** "

	mv db*.prc ${prc_dir}/db-${build}.prc
	builtfiles="$builtfiles db-${build}.prc"
	date=`date`
	if [ -f ${hfile} ]; then
		echo "ABORTING, the file ${hfile} must be moved out of the way"
		exit
	fi
	if [ -f ${bfile} ]; then
		echo "ABORTING, the file ${bfile} must be moved out of the way"
		exit
	fi


	# record the configuration options in BuildInfo file along with the prcs
	echo -e "\n\n---     =================     ---\ndb-${build} was built on $date with:\n" >> ${hfile}

	if [ -f ${prc_dir}/BuildInfo ]; then
        	cat ${prc_dir}/BuildInfo ${hfile} src/configuration > ${bfile}
        	mv ${bfile} ${prc_dir}/BuildInfo
	else
		cat ${hfile} src/configuration > ${prc_dir}/BuildInfo
   	fi
	rm ${hfile}
done


make uninstall
[ $? == 0 ] || die " **  Build Failed!  (uninstall) ** "
mv uninstall.prc ${prc_dir}/

for f in ${files}; do
	cp -r ${f} ${prc_dir}
done

# Create the actual snapshot prc zip file
zip -r ${prc_zip} ${prc_dir}
[ $? == 0 ] || die "couldn't make prc zip file."
rm -r ${prc_dir}


# Create the plugin zip file
make all-plugins
[ $? == 0 ] || die "  ** Build Failed!  (plugins) **"
find plugins -name "*.prc" | xargs -n1 -iqq mv qq ${plugin_dir}/
zip -r ${plugin_zip} ${plugin_dir}
[ $? == 0 ] || die "couldn't make plugin zip file."
rm -r ${plugin_dir}

# Clean up
make distclean

# Create the source zip file
dir=`pwd | sed -e "s/.*\/\([^\/]*\)$/\1/"`
cd ..
zip -r ${src_zip} ${dir}
[ $? == 0 ] || die "couldn't make source zip file."

