/*
 * DB: AddressDB Interface
 * Copyright (C) 2002-2003 by Marc Chalain (marc-chalain@users.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <PalmOS.h>

#include "../db.h"
#include "../io.h"
#include "../enum.h"
#include "../section.h"

#define TypeID 'DATA'
#define CreatorID 'addr'

typedef union
{
	struct
	{
		unsigned reserved	:13;
		unsigned note		:1;	// set if record contains a note handle
		unsigned custom4	:1;	// set if record contains a custom4
		unsigned custom3	:1;	// set if record contains a custom3
		unsigned custom2	:1;	// set if record contains a custom2
		unsigned custom1	:1;	// set if record contains a custom1
		unsigned title		:1;	// set if record contains a title
		unsigned country	:1;	// set if record contains a birthday
		unsigned zipCode	:1;	// set if record contains a birthday
		unsigned state		:1;	// set if record contains a birthday
		unsigned city		:1;	// set if record contains a birthday
		unsigned address	:1;	// set if record contains a address
		unsigned phone5		:1;	// set if record contains a phone5
		unsigned phone4		:1;	// set if record contains a phone4
		unsigned phone3		:1;	// set if record contains a phone3
		unsigned phone2		:1;	// set if record contains a phone2
		unsigned phone1		:1;	// set if record contains a phone1
		unsigned company	:1;	// set if record contains a company
		unsigned firstName	:1;	// set if record contains a firstName
		unsigned name		:1;	// set if record contains a name (bit 0)

	} bits;
	UInt32 allBits;
} AddrDBRecordFlags;

typedef union
{
	struct
	{
		unsigned reserved:8;
		unsigned displayPhoneForList:4;	// The phone displayed for the list view 0 - 4
		unsigned phone5:4;				// Which phone (home, work, car, ...)
		unsigned phone4:4;
		unsigned phone3:4;
		unsigned phone2:4;
		unsigned phone1:4;
	} phones;
	UInt32 phoneBits;
} AddrOptionsType;

typedef struct {
	AddrOptionsType		options;        // Display by company or by name
	AddrDBRecordFlags	flags;
	UInt8				companyFieldOffset;   // Offset from firstField
	char				firstField;
} PrvAddrPackedDBRecord;

struct private_data {
    UInt16 active_view;
    UInt16 top_visible_record;
    Boolean lfind_case_sensitive;
    Boolean lfind_whole_word;
};

static void Add_Close(DataSource * source) SECT_DRV;
static Boolean Add_Seek(DataSource * source, UInt16 * indexP, Int16 offset,Int16 direction) SECT_DRV;
static char * Add_Getter_GetString(void * data, UInt16 fieldNum) SECT_DRV;
static char * Add_Getter_GetNoteTitle(void * data, UInt16 fieldNum) SECT_DRV;
static char * Add_Getter_GetNoteTitle(void * data, UInt16 fieldNum) SECT_DRV;
static void Add_UnpackRecord(DataSource * source, void * packed, void * unpacked[]) SECT_DRV;
static void Add_LockRecord(DataSource * source, UInt16 index, Boolean writable, DataSourceGetter * getter, void ** data_ptr) SECT_DRV; 
static void Add_UnlockRecord(DataSource * source, UInt16 index, Boolean writable, void * record_data) SECT_DRV;
static Int16 Add_GetCurrentFilter(DataSource *) SECT_DRV;
static UInt8  Add_SupportsFeature(DataSource *source, FeatureType f) SECT_DRV;
static UInt16 Add_GetNumFeatures(DataSource * source, FeatureType f) SECT_DRV;
static Boolean Add_GetOption_Boolean(DataSource * source, UInt16 optionID) SECT_DRV;
static void Add_SetOption_Boolean(DataSource * source, UInt16 optionID, Boolean value) SECT_DRV;
static UInt16 Add_GetOption_UInt16(DataSource * source, UInt16 optionID) SECT_DRV;
static void Add_SetOption_UInt16(DataSource * source, UInt16 optionID, UInt16 value) SECT_DRV;
static Boolean Add_Capable(DataSource * source, DataSourceCapability capability) SECT_DRV;
static Err Add_OpenPDB(DataSource * source, UInt16 mode, UInt16 cardNo, LocalID dbID) SECT_DRV;
static Err Add_Open(DataSource* source, UInt16 mode, void* key_data, UInt16 key_size) SECT_DRV;
static Boolean Add_SupportsFieldType(DataSourceFieldType type) SECT_DRV;
static void Add_Enumerate(DataSourceEnumerationFunc callback, void * arg) SECT_DRV;
static Boolean Add_IsPresent(void * key_data, UInt32 key_size) SECT_DRV;
static Boolean Add_CanHandlePDB(UInt16 cardNo, LocalID dbID, UInt32 type, UInt32 creator) SECT_DRV;
static void Add_Delete(void * key_data, UInt32 key_size) SECT_DRV;


static void
Add_Close(DataSource * source)
{
    DmCloseDatabase(source->db);
    source->db = 0;
    DeallocateMemPtr(source->schema.fields);
    if (source->data) DeallocateMemPtr(source->data);
    if (source->key.data) DeallocateMemPtr(source->key.data);
    source->schema.numFields = 0;
    source->schema.fields = 0;
}

static Boolean
Add_Seek(DataSource * source, UInt16 * indexP, Int16 offset,
	    Int16 direction)
{
    if (DmSeekRecordInCategory(source->db, indexP, offset, direction,
                           dmAllCategories) != errNone) {
        return false;
    }
 
    return true;
}


static char *
Add_Getter_GetString(void * data, UInt16 fieldNum)
{
    void * * ptrs = data + sizeof(MemHandle);

    if (ptrs[fieldNum]) {
        return ptrs[fieldNum];
    } else
        return "";
}
 
static char *
Add_Getter_GetNoteTitle(void * data, UInt16 fieldNum)
{
    void * * ptrs = data + sizeof(MemHandle);

    if (ptrs[fieldNum])
        return ptrs[fieldNum];
    else
        return "";
}

static char *
Add_Getter_GetNote(void * data, UInt16 fieldNum)
{
    void * * ptrs = data + sizeof(MemHandle);

    if (ptrs[fieldNum])
        return ptrs[fieldNum];
    else
        return "";
}

static void
Add_UnpackRecord(DataSource * source, void * packed, void * unpacked[])
{
	UInt32 fieldFlags;
    PrvAddrPackedDBRecord * src = packed;
    UInt8 * rec = &src->firstField;
    UInt16 i;

    unpacked[0] = MemPtrRecoverHandle(packed);
    fieldFlags = src->flags.allBits;
    for (i = 0; i < source->schema.numFields; i++) {
        if ((fieldFlags >> i) & ((UInt32)0x00000001)) {
            unpacked[i + 1] = rec;
            rec += StrLen((char *)rec) + 1;
        } else {
            unpacked[i + 1] = 0;
        }
    }
}

static void
Add_LockRecord(DataSource * source, UInt16 index, Boolean writable,
                 DataSourceGetter * getter, void ** data_ptr)
{
    void ** ptrs, *packed;
    MemHandle h;

    ptrs = AllocateMemPtr(sizeof(MemHandle) + source->schema.numFields * sizeof(MemHandle));
 
    h = DmQueryRecord(source->db, index);

    packed = LockMemHandle(h);
    Add_UnpackRecord(source, packed, ptrs);
 
    getter->GetString  = Add_Getter_GetString;
    getter->GetNote    = Add_Getter_GetNote;
    getter->GetNoteTitle = Add_Getter_GetNoteTitle;
    *data_ptr = ptrs;
}

static void
Add_UnlockRecord(DataSource * source, UInt16 index, Boolean writable,
		    void * record_data)
{
    void * * ptrs = record_data;
 
    /*try to unlock the handler from the database record*/
    UnlockMemHandle(ptrs[0]);

    DeallocateMemPtr(ptrs);
}



static Int16
Add_GetCurrentFilter(DataSource * source)
{
	return dbNoFilter;
}

static UInt8 
Add_SupportsFeature(DataSource *source, FeatureType f) {
	return 0;
}

static UInt16
Add_GetNumFeatures(DataSource * source, FeatureType f) {
	return 0;
}

static Boolean
Add_GetOption_Boolean(DataSource * source, UInt16 optionID)
{
    struct private_data * data = (struct private_data *) source->data;
    Boolean result;

    switch (optionID) {
    case optionID_DisableGlobalFind:
    	result = true;
	break;

    case optionID_LFind_CaseSensitive:
        result = data->lfind_case_sensitive;
        break;

    case optionID_LFind_WholeWord:
        result = data->lfind_whole_word;
        break;

    default:
        result = false;
        break;
    }

    return result;
}

static void
Add_SetOption_Boolean(DataSource * source, UInt16 optionID, Boolean value)
{
    struct private_data * data = (struct private_data *) source->data;

    switch (optionID) {
    case optionID_DisableGlobalFind:
	break;

    case optionID_LFind_CaseSensitive:
        data->lfind_case_sensitive = value;
        break;

    case optionID_LFind_WholeWord:
        data->lfind_whole_word = value;
        break;
    }
}

static UInt16
Add_GetOption_UInt16(DataSource * source, UInt16 optionID)
{
    struct private_data * data = (struct private_data *) source->data;
    UInt16 result;

    switch (optionID) {
    case optionID_ListView_ActiveView:
        result = 0xFF;
        break;

    case optionID_ListView_TopVisibleRecord:
        result = data->top_visible_record;
        break;

    default:
        result = 0;
        break;
    }

    return result;
}

static void
Add_SetOption_UInt16(DataSource * source, UInt16 optionID, UInt16 value)
{
    struct private_data * data = (struct private_data *) source->data;

    switch (optionID) {
    case optionID_ListView_ActiveView:
    break;

    case optionID_ListView_TopVisibleRecord:
        data->top_visible_record = value;
    break;

    default:
    break;
    }
}

static Boolean
Add_Capable(DataSource * source, DataSourceCapability capability)
{
    return false;
}

static Err
Add_OpenPDB(DataSource * source, UInt16 mode, UInt16 cardNo, LocalID dbID)
{
    UInt32 dbType, size;
    UInt16 i;
    UInt8* p;

    MemSet(source, sizeof(DataSource), 0);

    /* Fill in the operations structure. */
    source->ops.Close = Add_Close;
    source->ops.Seek = Add_Seek;
    source->ops.LockRecord = Add_LockRecord;
    source->ops.UnlockRecord = Add_UnlockRecord;
	source->ops.RecalculateFields = NULL;
    source->ops.UpdateRecord = NULL;
    source->ops.SetField = NULL;
    source->ops.DeleteRecord = NULL;
    source->ops.GetRecordID = DS_GetRecordID;
    source->ops.GetRecordIndex = DS_GetRecordIndex;
	source->ops.GetCurrentFilter = Add_GetCurrentFilter;
    source->ops.GetNumFeatures = Add_GetNumFeatures;
    source->ops.GetFeature = NULL;
    source->ops.PutFeature = NULL;
    source->ops.StoreFeature = NULL;
    source->ops.DeleteFeature = NULL;
	source->ops.SupportsFeature = Add_SupportsFeature;
    source->ops.CheckUpdateSchema = NULL;
    source->ops.UpdateSchema = NULL;
    source->ops.GetOption_Boolean = Add_GetOption_Boolean;
    source->ops.SetOption_Boolean = Add_SetOption_Boolean;
    source->ops.GetOption_UInt16 = Add_GetOption_UInt16;
    source->ops.SetOption_UInt16 = Add_SetOption_UInt16;
    source->ops.GetKey = DS_GetKey;
    source->ops.GetPDB = DS_PDB_GetPDB;
    source->ops.GetTitle = DS_PDB_GetTitle;
    source->ops.SetTitle = DS_PDB_SetTitle;
    source->ops.Sort = NULL;
    source->ops.Capable = Add_Capable;
    source->ops.Info = DS_PDB_Info;
    source->ops.RecalculateFields = NULL;

    source->currentcategory = dmAllCategories;

    /* Private data to emulate some options supported other drivers. */
    source->data = AllocateMemPtr(sizeof(struct private_data));
    ((struct private_data *) source->data)->active_view = 0;
    ((struct private_data *) source->data)->top_visible_record = 0;
    ((struct private_data *) source->data)->lfind_case_sensitive = false;
    ((struct private_data *) source->data)->lfind_whole_word = false;

    /* Check to see if we support this type of database. */
    DmDatabaseInfo(cardNo, dbID, 0, 0, 0, 0, 0, 0, 0, 0, 0, &dbType, 0);
    if (dbType != TypeID)
        return dmErrCantOpen;

    /* Open this database. */
    source->db = DmOpenDatabase(cardNo, dbID, mode);
    if (! source->db)
        return DmGetLastErr();

    /* The number of fields from the header. */
    source->schema.numFields = 19;

    /* Determine how much space needs to be allocated for schema fields. */
    size = source->schema.numFields * sizeof(DataSourceFieldInfo);
    for (i = 0; i < source->schema.numFields; ++i) {
        size += 33;
    }
    source->schema.fields = AllocateMemPtr(size);
    MemSet(source->schema.fields, size, 0);

    /* Extract the entire schema from the database header. */
    p = (UInt8 *) (source->schema.fields + source->schema.numFields);
    for (i = 0; i < source->schema.numFields; ++i) {
        if (i < source->schema.numFields - 1)
           source->schema.fields[i].type = FIELD_TYPE_STRING;
        else
            source->schema.fields[i].type = FIELD_TYPE_NOTE;

        /* Copy the field name into the schema and point schema at it. */
        SysStringByIndex(strt_FieldNames, i, p, 32);
        source->schema.fields[i].name = (char *)p;
        p += 33;
    }

    return 0;
}

static Err
Add_Open(DataSource* source, UInt16 mode, void* key_data, UInt16 key_size)
{
    Err error;
    DataSourceChooserKey * key = (DataSourceChooserKey *) key_data;
    LocalID dbID;

    mode = dmModeReadOnly;

    /* Make sure that the key has a valid size. */
    if (key_size != sizeof(DataSourceChooserKey))
        return dmErrCantOpen;

    /* Retrieve the location in memory of this database. */
    dbID = DmFindDatabase(key->cardNo, key->name);
    if (!dbID)
        return DmGetLastErr();

    /* Let the PDB open code handle the remainder of the open. */
    error = Add_OpenPDB(source, mode & ~MaskReadEdit, key->cardNo, dbID);
    if (!error) {
        /* store the key of the database*/
        source->key.size = key_size;
        source->key.data = AllocateMemPtr(key_size);
        source->openMode = mode;
        MemMove( source->key.data, key_data, key_size);
    }
    return error;
}

static Boolean
Add_SupportsFieldType(DataSourceFieldType type)
{
    switch (type) {
    case FIELD_TYPE_STRING:
    case FIELD_TYPE_NOTE:
        return true;
    default:
        return false;
    }
}

static void
Add_Enumerate(DataSourceEnumerationFunc callback, void * arg)
{
    DS_Enumerate(callback, arg, TypeID, CreatorID, sourceID_Add);
}

static Boolean
Add_IsPresent(void * key_data, UInt32 key_size)
{
    return DS_IsPresent(key_data, key_size, TypeID, CreatorID);
}

static Boolean
Add_CanHandlePDB(UInt16 cardNo, LocalID dbID, UInt32 type, UInt32 creator)
{
    return (creator == CreatorID) && (type == TypeID);
}

static void
Add_Delete(void * key_data, UInt32 key_size)
{
}

DataSourceDriver *
Add_GetDriver(void)
{
    DataSourceDriver * driver = AllocateMemPtr(sizeof(DataSourceDriver));

    StrCopy(driver->name, "AddressDB");
    driver->sourceID = sourceID_Add;
    driver->class_ops.Open = Add_Open;
    driver->class_ops.OpenPDB = Add_OpenPDB;
    driver->class_ops.CheckTitle = DS_CheckTitle;
    driver->class_ops.Create = NULL;
    driver->class_ops.SupportsFieldType = Add_SupportsFieldType;
    driver->class_ops.Delete = Add_Delete;
    driver->class_ops.Enumerate = Add_Enumerate;
    driver->class_ops.GlobalFind = NULL;
    driver->class_ops.IsPresent = Add_IsPresent;
    driver->class_ops.CompareKey = DS_CompareKey;
    driver->class_ops.CanHandlePDB = Add_CanHandlePDB;
    driver->class_ops.Exchange = DS_PDB_Exchange;
	driver->class_ops.CompileScript = NULL;

	driver->next = NULL;
    return driver;
}
